#define		SUMMARY_INTERVAL	300000

#define		MAXTIMERS		32

#define		TMR_IDLE		0
#define		TMR_ARTWRITE		1
#define		TMR_ARTLINK		2
#define		TMR_HISWRITE		3
#define		TMR_HISSYNC		4
#define		TMR_SITESEND		5
#define		TMR_ARTCTRL		6
#define		TMR_ARTCNCL		7
#define		TMR_HISHAVE		8
#define		TMR_HISGREP		9

extern void TMRinit();
extern void TMRmainloophook();
extern void TMRstart(int);
extern void TMRstop(int);
